import java.net.URI
import java.net.http.HttpClient
import java.net.http.HttpRequest
import java.net.http.HttpResponse

import kotlinx.serialization.*
import kotlinx.serialization.json.*

const val froomleApi = "https://{subdomain}.froomle.com/api/{environment}"

val jsonFormat = Json { explicitNulls = false; ignoreUnknownKeys = true }

@Serializable
data class Events(
    val events: List<Event>
)

@Serializable
data class Event(
    val event_type: String,
    val channel: String,
    val page_type: String,
    val device_id: String,
    val user_id: String? = null,
    val action_item: String,
    val action_item_type: String,

    val request_id: String? = null,
    val user_group: String? = null,
    val list_name: String? = null,
)

@Serializable
data class RecommendationRequest(
    val channel: String,
    val page_type: String,
    val context_item: String? = null,
    val context_item_type: String? = null,
    val device_id: String,
    val user_id: String? = null,
    val lists: List<RequestList>,
)

@Serializable
data class RequestList(
    val list_name: String,
    val list_size: Int,
    val configuration_id: String? = null,
)

@Serializable
data class Recommendations(
    val request_id: String,
    val user_group: String,
    val lists: List<ResponseList>,
)

@Serializable
data class ResponseList(
    val list_name: String,
    val list_size: Int,
    val configuration_id: String? = null,
    val items: List<Item>,
)

@Serializable
data class Item(
    val item_id: String,
    val rank: Int,
)

fun sendEvents(events: Events) {
    val client = HttpClient.newBuilder().build()

    val body = jsonFormat.encodeToString(events)
    val httpRequest = HttpRequest.newBuilder()
        .uri(URI.create("$froomleApi/events"))
        .header("Content-Type", "application/json")
        .POST(HttpRequest.BodyPublishers.ofString(body))
        .build()
    val httpResponse = client.send(httpRequest, HttpResponse.BodyHandlers.ofString())

    if (httpResponse.statusCode() >= 300) {
        throw Exception("Unexpected ${httpResponse.statusCode()}: ${httpResponse.body()}")
    }
}

fun fetchRecommendations(recommendationRequest: RecommendationRequest): Recommendations {
    val client = HttpClient.newBuilder().build()

    val body = jsonFormat.encodeToString(recommendationRequest)
    val httpRequest = HttpRequest.newBuilder()
        .uri(URI.create("$froomleApi/recommendations/requests"))
        .header("Content-Type", "application/json")
        .POST(HttpRequest.BodyPublishers.ofString(body))
        .build()
    val httpResponse = client.send(httpRequest, HttpResponse.BodyHandlers.ofString())

    if (httpResponse.statusCode() >= 300) {
        throw Exception("Unexpected ${httpResponse.statusCode()}: ${httpResponse.body()}")
    }

    return jsonFormat.decodeFromString(httpResponse.body())
}

fun main(args: Array<String>) {

    // Contextual information about where the recommendations are served

    val channel = "..."
    val page_type = "..."
    val context_item = "..."       // null if page is not an article/product detail page
    val context_item_type = "..."  // null if page is not an article/product detail page
    val device_id = "..."
    val user_id = "..."            // null for non-logged-in users

    val list_name = "..."

    // Request recommendations and display the recommended items

    val recommendationRequest = RecommendationRequest(
        channel = channel,
        page_type = page_type,
        context_item = context_item,
        context_item_type = context_item_type,
        device_id = device_id,
        user_id = user_id,
        lists = listOf(
            RequestList(
                list_name = list_name,
                list_size = 5,
            )
        ),
    )
    val recommendations = fetchRecommendations(recommendationRequest)
    val responseList = recommendations.lists[0]
    for (item in responseList.items) {
        println("Recommended item ${item.rank}: ${item.item_id}")
    }

    // Sending impression events for recommended items (ideally only once user sees them)

    val impressionEvents = Events(events = responseList.items.map {
        Event(
            event_type = "impression",
            channel = channel,
            page_type = page_type,
            device_id = device_id,
            user_id = user_id,
            action_item = it.item_id,
            action_item_type = "article",

            request_id = recommendations.request_id,
            user_group = recommendations.user_group,
            list_name = list_name,
        )
    })
    sendEvents(impressionEvents)

    // Sending a click_on_recommendation event for when a recommendation is clicked

    val clickedItem = responseList.items[0].item_id  // assuming the 1st recommendation was clicked

    val clickOnRecommendationEvents = Events(
        events = listOf(
            Event(
                event_type = "click_on_recommendation",
                channel = channel,
                page_type = page_type,
                device_id = device_id,
                user_id = user_id,
                action_item = clickedItem,
                action_item_type = "article",

                request_id = recommendations.request_id,
                user_group = recommendations.user_group,
                list_name = list_name,
            )
        )
    )
    sendEvents(clickOnRecommendationEvents)

    // Sending a detail_pageview event for when a product/article is viewed

    val viewedItem = "..."

    val detailPageviewEvents = Events(
        events = listOf(
            Event(
                event_type = "detail_pageview",
                channel = channel,
                page_type = page_type,
                device_id = device_id,
                user_id = user_id,
                action_item = viewedItem,
                action_item_type = "article",
            )
        )
    )
    sendEvents(detailPageviewEvents)
}
